#!/usr/bin/env python
# coding: utf-8
"""Experimental function to compute the Response Amplitude Operator."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging

import numpy as np
import xarray as xr

LOG = logging.getLogger(__name__)


def rao(dataset, wave_direction=0.0, dissipation=None, stiffness=None):
    """Response Amplitude Operator.

    Parameters
    ----------
    dataset: xarray Dataset
        The hydrodynamical dataset.
        This function supposes that variables named 'mass' and 'hydrostatic_stiffness' are in the dataset.
        Other variables can be computed by Capytaine, by those two should be manually added to the dataset.
    wave_direction: float, optional
        The direction of the incoming waves.
        Default: 0 radian (x direction)
    dissipation: array, optional
        An optional dissipation matrix (e.g. Power Take Off) to be included in the RAO.
        Default: none.
    stiffness: array, optional
        An optional stiffness matrix (e.g. mooring stiffness) to be included in the RAO.
        Default: none.

    Returns
    -------
    xarray DataArray
        The RAO as an array depending of omega and the degree of freedom.
    """
    LOG.info("Compute RAO.")

    # ASSEMBLE MATRICES
    omega = dataset.coords['omega']  # Range of frequencies in the dataset

    A = (-omega**2*(dataset['mass'] + dataset['added_mass'])
         + 1j*omega*dataset['radiation_damping']
         + dataset['hydrostatic_stiffness'])

    if dissipation is not None:
        A += 1j*omega*dissipation

    if stiffness is not None:
        A += stiffness

    if 'excitation_force' not in dataset:
        dataset['excitation_force'] = dataset['Froude_Krylov_force'] + dataset['diffraction_force']
    excitation = dataset['excitation_force'].sel(wave_direction=wave_direction)

    # SOLVE LINEAR SYSTEMS
    # Reorder dimensions of the arrays to be sure to solve the right system.
    A = A.transpose('omega', 'radiating_dof', 'influenced_dof')
    excitation = excitation.transpose('omega',  'influenced_dof')

    # Solve the linear systems (one for each value of omega)
    X = np.linalg.solve(A, excitation)

    return xr.DataArray(X, coords=[omega, dataset.coords['radiating_dof']], dims=['omega', 'radiating_dof'])




#!/usr/bin/env python
# coding: utf-8
"""VTK animation for the free surface elevation."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging

import numpy as np
from numpy import pi
import vtk

from capytaine.ui.vtk.helpers import compute_node_data, compute_vtk_polydata

LOG = logging.getLogger(__name__)


class Animation:
    """Class to generate an animation of a result of Capytaine,
    including the elevation of the free surface.

    The animation is made of a short loop of a single period of the solution in frequency domain.

    Parameters
    ----------
    loop_duration: float
        Duration in the loop. For real time animation, the period of the motion.
    fps: int, optional
        Number of frames per second in the animation (default: 24).

    Attributes
    ----------
    frames_per_loop: int
        Number of frames in one loop
    actors: list of vtk actor objects
        The objects in the scene.
    """

    def __init__(self, loop_duration, fps=24):
        self.fps = fps
        self.frames_per_loop = int(fps * loop_duration)
        self.actors = []

        self._precomputed_polydatas = {}
        self._current_frame = 0

    # @classmethod
    # def from_result(self, result):
    #     from capytaine.bodies import TRANSLATION_DOFS_DIRECTIONS, ROTATION_DOFS_AXIS
    #
    #         if display_dof.lower() in TRANSLATION_DOFS_DIRECTIONS:
    #             direction = np.asarray(TRANSLATION_DOFS_DIRECTIONS[display_dof.lower()])
    #             def translation_motion(self, frame):
    #                 nonlocal direction
    #                 pos = np.asarray(self.body_actor.GetPosition())
    #                 pos = (1 - direction) * pos + \
    #                       direction * np.cos(2*np.pi*(frame % self.frames_per_loop)/self.frames_per_loop)
    #                 self.body_actor.SetPosition(*pos)
    #             self.update_body_position = translation_motion
    #
    #         elif display_dof.lower() in ROTATION_DOFS_AXIS:
    #             direction = np.asarray(ROTATION_DOFS_AXIS[display_dof.lower()])
    #             def rotation_motion(self, frame):
    #                 nonlocal direction
    #                 pos = np.asarray(self.body_actor.GetOrientation())
    #                 pos = (1 - direction) * pos + \
    #                       direction * np.cos(2*np.pi*(frame % self.frames_per_loop)/self.frames_per_loop)
    #                 self.body_actor.SetOrientation(*pos)
    #             self.update_body_position = rotation_motion


    def _add_actor(self, mesh, faces_motion=None, faces_colors=None, edges=False):
        """Add an animated object to the scene."""
        if faces_motion is not None:
            nodes_motion = compute_node_data(mesh.merged(), faces_motion)
        else:
            nodes_motion = None

        base_polydata = compute_vtk_polydata(mesh.merged())
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputData(base_polydata)

        actor = vtk.vtkActor()
        if edges:
            actor.GetProperty().EdgeVisibilityOn()
        actor.GetProperty().SetInterpolationToGouraud()
        actor.SetMapper(mapper)

        if nodes_motion is not None or faces_colors is not None:
            LOG.info(f"Precompute motions of {mesh.name} before animation.")
            self._precomputed_polydatas[actor] = []

            for i_frame in range(self.frames_per_loop):
                new_polydata = vtk.vtkPolyData()
                new_polydata.DeepCopy(base_polydata)

                if nodes_motion is not None:
                    # Change points positions at frame i
                    current_deformation = (
                            np.abs(nodes_motion)*np.cos(np.angle(nodes_motion)-2*pi*i_frame/self.frames_per_loop)
                    )

                    points = new_polydata.GetPoints()
                    for j in range(mesh.nb_vertices):
                        point = points.GetPoint(j)
                        point = np.asarray(point) + current_deformation[j]
                        points.SetPoint(j, tuple(point))

                if faces_colors is not None:
                    # Evaluate scalar field at frame i
                    current_colors = (
                        np.abs(faces_colors)*np.cos(np.angle(faces_colors)-2*pi*i_frame/self.frames_per_loop)
                    )
                    max_val = max(abs(faces_colors))
                    vtk_faces_colors = vtk.vtkFloatArray()
                    for i, color in enumerate(current_colors):
                        vtk_faces_colors.InsertValue(i, (color+max_val)/(2*max_val))
                    new_polydata.GetCellData().SetScalars(vtk_faces_colors)

                self._precomputed_polydatas[actor].append(new_polydata)
        else:
            self._precomputed_polydatas[actor] = None

        self.actors.append(actor)

        return actor

    def add_body(self, body, faces_motion=None, faces_colors=None, edges=False):
        """Add an floating body to the scene.

        Parameters
        ----------
        body: FloatingBody
            The object to include in the scene.
        faces_motion: dof, optional
            The motion of the body defined at the center of the faces.
        faces_colors: iterable of complex numbers, optional
            Scalar field over the surface of the body that should be displayed with colors.
        edges: bool, optional
            Draw the edges of the mesh in the scene.

        Returns
        -------
        vtk actor object
        """
        actor = self._add_actor(body.mesh.merged(), faces_motion=faces_motion,
                                faces_colors=faces_colors, edges=edges)
        if faces_colors is None:
            actor.GetProperty().SetColor((1, 1, 0))
        else:
            lut = vtk.vtkLookupTable()
            lut.SetNumberOfColors(50)
            lut.SetHueRange(0, 0.6)
            lut.SetSaturationRange(0.5, 0.5)
            lut.SetValueRange(0.8, 0.8)
            lut.Build()
            actor.GetMapper().SetLookupTable(lut)

        return actor

    def add_free_surface(self, free_surface, faces_elevation):
        """Add the free surface to the scene.

        Parameters
        ----------
        free_surface: FreeSurface
            The free surface object
        faces_elevation: array of complex numbers
            The elevation of each face of the meshed free surface given as a complex number.

        Returns
        -------
        vtk actor object
        """
        faces_motion = np.array([(0, 0, elevation) for elevation in faces_elevation])
        actor = self._add_actor(free_surface.mesh, faces_motion=faces_motion, faces_colors=faces_motion[:, 2])

        lut = vtk.vtkLookupTable()
        lut.SetNumberOfColors(50)
        lut.SetHueRange(0.58, 0.58)
        lut.SetSaturationRange(0.5, 0.5)
        lut.SetValueRange(0.4, 0.6)
        lut.Build()
        actor.GetMapper().SetLookupTable(lut)

        return actor

    def _callback(self, renderer, event):
        for actor in self.actors:
            if self._precomputed_polydatas[actor] is not None:
                actor.GetMapper().SetInputData(self._precomputed_polydatas[actor][self._current_frame % self.frames_per_loop])
        renderer.GetRenderWindow().Render()
        self._current_frame += 1

    def run(self, camera_position=(-10.0, -10.0, 10.0), resolution=(1280, 720)):
        """Run the animation.

        Parameters
        ----------
        camera_position: 3-ple of floats, optional
            The starting position of the camera in the scene.
        resolution: 2-ple of ints, optional
            Resolution of the video in pixels.
        """
        # Setup a renderer, render window, and interactor
        renderer = vtk.vtkRenderer()
        renderer.SetBackground(1, 1, 1)  # Background color white
        for actor in self.actors:
            renderer.AddActor(actor)
        renderer.Modified()

        camera = vtk.vtkCamera()
        camera.SetPosition(*camera_position)
        camera.SetFocalPoint(0, 0, 0)
        camera.SetViewUp(0, 0, 1)
        renderer.SetActiveCamera(camera)

        render_window = vtk.vtkRenderWindow()
        render_window.SetSize(*resolution)
        render_window.SetWindowName("Capytaine animation")
        render_window.AddRenderer(renderer)

        render_window_interactor = vtk.vtkRenderWindowInteractor()
        render_window_interactor.SetRenderWindow(render_window)
        render_window_interactor.GetInteractorStyle().SetCurrentStyleToTrackballCamera()

        render_window.Render()

        render_window_interactor.Initialize()  # Initialize must be called prior to creating timer events.

        render_window_interactor.AddObserver('TimerEvent', self._callback)
        render_window_interactor.CreateRepeatingTimer(int(1000 / self.fps))

        render_window_interactor.Start()

        # Run until stopped by user.

        del render_window_interactor
        del render_window

    def save(self, filepath, nb_loops=1, camera_position=(-10.0, -10.0, 10.0), resolution=(1280, 720)):
        """Save the animation in a video file.

        Parameters
        ----------
        filepath: string
            Path of the output file.
        nb_loop: int, optional
            Number of periods to save in the file.
        camera_position: 3-ple of floats, optional
            The starting position of the camera in the scene.
        resolution: 2-ple of ints, optional
            Resolution of the video in pixels.
        """
        renderer = vtk.vtkRenderer()
        renderer.SetBackground(1, 1, 1)  # Background color white
        for actor in self.actors:
            renderer.AddActor(actor)
        renderer.Modified()

        camera = vtk.vtkCamera()
        camera.SetPosition(*camera_position)
        camera.SetFocalPoint(0, 0, 0)
        camera.SetViewUp(0, 0, 1)
        renderer.SetActiveCamera(camera)

        render_window = vtk.vtkRenderWindow()
        render_window.SetSize(*resolution)
        render_window.OffScreenRenderingOn()
        render_window.AddRenderer(renderer)

        image_filter = vtk.vtkWindowToImageFilter()
        image_filter.SetInput(render_window)
        image_filter.SetInputBufferTypeToRGB()
        image_filter.ReadFrontBufferOff()

        writer = vtk.vtkOggTheoraWriter()
        writer.SetInputConnection(image_filter.GetOutputPort())
        writer.SetFileName(filepath)
        writer.SetRate(self.fps)

        writer.Start()

        for i_frame in range(nb_loops*self.frames_per_loop):
            self._callback(renderer, None)
            image_filter.Modified()
            writer.Write()

        writer.End()
        render_window.Finalize()

        del image_filter
        del writer
        del render_window



#!/usr/bin/env python
# coding: utf-8
"""Tools for mesh quality and mesh healing.
Based on meshmagick <https://github.com/LHEEA/meshmagick> by François Rongère.
"""
# Copyright (C) 2017-2019 Matthieu Ancellin, based on the work of François Rongère
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging

import numpy as np

from capytaine.meshes.geometry import inplace_transformation

LOG = logging.getLogger(__name__)


def merge_duplicates(mesh, atol=1e-8):
    """Merges the duplicate vertices of the mesh in place.

    Parameters
    ----------
    atol : float, optional
        Absolute tolerance. default is 1e-8

    Returns
    -------
    new_id : ndarray
        Array of indices that merges the vertices.
    """
    uniq, new_id = merge_duplicate_rows(mesh.vertices, atol=atol)

    nv_init = mesh.nb_vertices

    # Updating mesh data
    mesh.vertices = uniq
    mesh.faces = new_id[mesh.faces] # Faces vertices ids are updated here

    nv_final = mesh.nb_vertices

    LOG.debug("* Merging duplicate vertices that lie in an absolute proximity of %.1E...", atol)
    delta_n = nv_init - nv_final
    if delta_n == 0:
        LOG.debug("\t--> No duplicate vertices have been found")
    else:
        LOG.debug("\t--> Initial number of vertices : %u", nv_init)
        LOG.debug("\t--> Final number of vertices   : %u", nv_final)
        LOG.debug("\t--> %u vertices have been merged", delta_n)

    # if mesh._has_connectivity():
    #     mesh._remove_connectivity()

    return new_id


def merge_duplicate_rows(arr, atol=1e-8):
    """Returns a new node array where close nodes have been merged into one node (following atol).

    Parameters
    ----------
    arr : array_like
        array of the coordinates of the mesh's nodes
    atol : float, optional
        the tolerance used to define nodes that are coincident and
        that have to be merged

    Returns
    -------
    arr : ndarray
        array of the coordinates of the mesh's nodes where
        every node is different
    newID : ndarray
        array of the new new vertices IDs
    """
    # This function is a bottleneck in the clipping routines
    # TODO: use np.unique to cluster groups --> acceleration !!

    # atol = pow(10, -decimals)

    arr = np.asarray(arr)

    nv, nbdim = arr.shape

    levels = [0, nv]
    iperm = np.arange(nv)

    for dim in range(nbdim):
        # Sorting the first dimension
        values = arr[:, dim].copy()
        if dim > 0:
            values = values[iperm]
        levels_tmp = []
        for (ilevel, istart) in enumerate(levels[:-1]):
            istop = levels[ilevel+1]

            if istop-istart > 1:
                level_values = values[istart:istop]
                iperm_view = iperm[istart:istop]

                iperm_tmp = level_values.argsort()

                level_values[:] = level_values[iperm_tmp]
                iperm_view[:] = iperm_view[iperm_tmp]

                levels_tmp.append(istart)
                vref = values[istart]

                for idx in range(istart, istop):
                    cur_val = values[idx]
                    if np.abs(cur_val - vref) > atol:
                        levels_tmp.append(idx)
                        vref = cur_val

            else:
                levels_tmp.append(levels[ilevel])
        if len(levels_tmp) == nv:
            # No duplicate rows
            # if verbose:
            # LOG.debug "\t -> No duplicate _vertices detected :)"
            newID = np.arange(nv)

        levels_tmp.append(nv)
        levels = levels_tmp

    else:
        # Building the new merged node list
        arr_tmp = []
        newID = np.arange(nv)
        for (ilevel, istart) in enumerate(levels[:-1]):
            istop = levels[ilevel+1]

            arr_tmp.append(arr[iperm[istart]])
            newID[iperm[list(range(istart, istop))]] = ilevel
        arr = np.array(arr_tmp, dtype=float)
        # Applying renumbering to cells
        # if F is not None:
        #     for cell in F:
        #         cell[:] = newID[cell]

        # if verbose:
        # nv_new = arr.shape[0]
        # LOG.debug "\t -> Initial number of nodes : {:d}".format(nv)
        # LOG.debug "\t -> New number of nodes     : {:d}".format(nv_new)
        # LOG.debug "\t -> {:d} nodes have been merged".format(nv-nv_new)

    # if F is not None:
    #     if return_index:
    #         return arr, F, newID
    #     else:
    #         return arr, F
    # else:
    return arr, newID


@inplace_transformation
def heal_normals(mesh):
    """Heals the mesh's normals orientations so that they have a consistent orientation and try to make them outward.
    """
    # TODO: return the different groups of a mesh in case it is made of several unrelated groups

    nv = mesh.nb_vertices
    nf = mesh.nb_faces
    faces = mesh._faces

    # Building connectivities
    v_v = mesh.vv
    v_f = mesh.vf
    f_f = mesh.ff
    boundaries = mesh.boundaries

    if len(boundaries) > 0:
        mesh_closed = False
    else:
        mesh_closed = True

    # Flooding the mesh to find inconsistent normals
    type_cell = np.zeros(nf, dtype=np.int32)
    type_cell[:] = 4
    type_cell[mesh.triangles_ids] = 3

    f_vis = np.zeros(nf, dtype=bool)
    f_vis[0] = True
    stack = [0]
    nb_reversed = 0
    while 1:
        if len(stack) == 0:
            if np.any(np.logical_not(f_vis)):
                iface = np.where(np.logical_not(f_vis))[0][0]
                stack.append(iface)
                f_vis[iface] = True
            else:
                break

        iface = stack.pop()
        face = faces[iface]
        s1 = set(face)

        for iadj_f in f_f[iface]:
            if f_vis[iadj_f]:
                continue
            f_vis[iadj_f] = True
            # Removing the other pointer
            f_f[iadj_f].remove(iface)  # So as it won't go from iadj_f to iface in the future

            # Shared vertices
            adjface = faces[iadj_f]
            s2 = set(adjface)
            # try:
            common_vertices = list(s1 & s2)
            if len(common_vertices) == 2:
                i_v1, i_v2 = common_vertices
            else:
                LOG.warning('faces %u and %u have more than 2 vertices in common !', iface, iadj_f)
                continue

            # Checking normal consistency
            face_ref = np.roll(face[:type_cell[iface]], -np.where(face == i_v1)[0][0])
            adj_face_ref = np.roll(adjface[:type_cell[iadj_f]], -np.where(adjface == i_v1)[0][0])

            if face_ref[1] == i_v2:
                i = 1
            else:
                i = -1

            if adj_face_ref[i] == i_v2:
                # Reversing normal
                nb_reversed += 1
                faces[iadj_f] = np.flipud(faces[iadj_f])

            # Appending to the stack
            stack.append(iadj_f)

    LOG.debug("* Healing normals to make them consistent and if possible outward")
    if nb_reversed > 0:
        LOG.debug('\t--> %u faces have been reversed to make normals consistent across the mesh' % (nb_reversed))
    else:
        LOG.debug("\t--> Normals orientations are consistent")

    mesh._faces = faces

    # Checking if the normals are outward
    if mesh_closed:
        zmax = np.max(mesh._vertices[:, 2])

        areas = mesh.faces_areas
        normals = mesh.faces_normals
        centers = mesh.faces_centers
        # areas, normals, centers = get_all_faces_properties(vertices, faces)

        hs = (np.array([(centers[:, 2] - zmax) * areas, ] * 3).T * normals).sum(axis=0)

        tol = 1e-9
        if np.fabs(hs[0]) > tol or np.fabs(hs[1]) > tol:
            LOG.warning("\t--> the mesh does not seem watertight althought marked as closed...")

        if hs[2] < 0:
            flipped = True
            mesh.flip_normals()
        else:
            flipped = False

        if flipped:
            LOG.debug('\t--> Every normals have been reversed to be outward')

    else:
        LOG.info("\t--> Mesh is not closed, meshmagick cannot test if the normals are outward")

    return mesh


@inplace_transformation
def remove_unused_vertices(mesh):
    """Removes unused vertices in the mesh in place.

    Those are vertices that are not used by any face connectivity.
    """
    # TODO: implementer return_index !!
    nv = mesh.nb_vertices
    vertices, faces = mesh._vertices, mesh._faces

    used_v = np.zeros(nv, dtype=np.bool)
    used_v[sum(list(map(list, faces)), [])] = True
    nb_used_v = sum(used_v)

    if nb_used_v < nv:
        new_id__v = np.arange(nv)
        new_id__v[used_v] = np.arange(nb_used_v)
        faces = new_id__v[faces]
        vertices = vertices[used_v]

    mesh._vertices, mesh._faces = vertices, faces

    LOG.debug("* Removing unused vertices in the mesh:")
    if nb_used_v < nv:
        unused_v = np.where(np.logical_not(used_v))[0]
        vlist_str = '[' + ', '.join(str(iV) for iV in unused_v) + ']'
        LOG.debug("\t--> %u unused vertices have been removed" % (nv - nb_used_v))
    else:
        LOG.debug("\t--> No unused vertices")

    return mesh


@inplace_transformation
def heal_triangles(mesh):
    """Makes the triangle connectivity consistent (in place).

    A general face is stored internally as a 4 integer array. It allows to describe indices of a quadrangle's vertices. For triangles, the first index should be equal to the last. This method ensures that this rule is applied everywhere and correct bad triangles description.
    """
    faces = mesh._faces

    quads = faces[:, 0] != faces[:, -1]
    nquads_init = sum(quads)

    faces[quads] = np.roll(faces[quads], 1, axis=1)
    quads = faces[:, 0] != faces[:, -1]

    faces[quads] = np.roll(faces[quads], 1, axis=1)
    quads = faces[:, 0] != faces[:, -1]

    faces[quads] = np.roll(faces[quads], 1, axis=1)
    quads = faces[:, 0] != faces[:, -1]
    nquads_final = sum(quads)

    mesh._faces = faces

    LOG.debug("* Ensuring consistent definition of triangles:")
    if nquads_final < nquads_init:
        LOG.debug("\t--> %u triangles were described the wrong way and have been corrected" % (
        nquads_init - nquads_final))
    else:
        LOG.debug("\t--> Triangle description is consistent")

    return mesh


@inplace_transformation
def remove_degenerated_faces(mesh, rtol=1e-5):
    """Removes tiny triangles from the mesh (in place).

    Tiny triangles are those whose area is lower than the mean triangle area in the mesh times the relative
    tolerance given.

    Parameters
    ----------
    rtol : float, optional
        Positive relative tolerance
    """

    assert 0 < rtol

    # TODO: implementer un retour d'index des faces extraites
    areas = mesh.faces_areas
    area_threshold = areas.mean() * float(rtol)

    # Detecting faces that have null area
    faces = mesh._faces[np.logical_not(areas < area_threshold)]
    nb_removed = mesh.nb_faces - faces.shape[0]
    LOG.debug('* Removing degenerated faces')
    if nb_removed > 0:
        LOG.debug('\t-->%u degenerated faces have been removed' % nb_removed)
    else:
        LOG.debug('\t--> No degenerated faces')

    mesh._faces = faces

    return mesh


def print_quality(mesh):
    """Returns data on the mesh quality.
    Needs to be tested...

    It uses VTK and is reproduced from
    http://vtk.org/gitweb?p=VTK.git;a=blob;f=Filters/Verdict/Testing/Python/MeshQuality.py
    """
    # This function is reproduced from
    # http://vtk.org/gitweb?p=VTK.git;a=blob;f=Filters/Verdict/Testing/Python/MeshQuality.py
    polydata = mesh._vtk_polydata()
    quality = vtk.vtkMeshQuality()
    quality.SetInputData(polydata)

    def DumpQualityStats(iq, arrayname):
        an = iq.GetOutput().GetFieldData().GetArray(arrayname)
        cardinality = an.GetComponent(0, 4)
        range = list()
        range.append(an.GetComponent(0, 0))
        range.append(an.GetComponent(0, 2))
        average = an.GetComponent(0, 1)
        stdDev = math.sqrt(math.fabs(an.GetComponent(0, 3)))
        outStr = '%s%g%s%g\n%s%g%s%g' % (
            '    range: ', range[0], '  -  ', range[1],
            '    average: ', average, '  , standard deviation: ', stdDev)
        return outStr

    # Here we define the various mesh types and labels for output.
    meshTypes = [
        ['Triangle', 'Triangle',
         [['QualityMeasureToArea', ' Area Ratio:'],
          ['QualityMeasureToEdgeRatio', ' Edge Ratio:'],
          ['QualityMeasureToAspectRatio', ' Aspect Ratio:'],
          ['QualityMeasureToRadiusRatio', ' Radius Ratio:'],
          ['QualityMeasureToAspectFrobenius', ' Frobenius Norm:'],
          ['QualityMeasureToMinAngle', ' Minimal Angle:']
          ]
         ],

        ['Quad', 'Quadrilateral',
         [['QualityMeasureToArea', ' Area Ratio:'],
          ['QualityMeasureToEdgeRatio', ' Edge Ratio:'],
          ['QualityMeasureToAspectRatio', ' Aspect Ratio:'],
          ['QualityMeasureToRadiusRatio', ' Radius Ratio:'],
          ['QualityMeasureToMedAspectFrobenius',
           ' Average Frobenius Norm:'],
          ['QualityMeasureToMaxAspectFrobenius',
           ' Maximal Frobenius Norm:'],
          ['QualityMeasureToMinAngle', ' Minimal Angle:']
          ]
         ]
    ]
    res = ''
    if polydata.GetNumberOfCells() > 0:
        for meshType in meshTypes:
            res += '\n%s%s' % (meshType[1], ' quality of the mesh ')
            quality.Update()
            an = quality.GetOutput().GetFieldData().GetArray('Mesh ' + meshType[1] + ' Quality')
            cardinality = an.GetComponent(0, 4)

            res = ''.join((res, '(%u elements):\n' % cardinality))

            # res += '('+str(cardinality) +meshType[1]+'):\n'

            for measure in meshType[2]:
                eval('quality.Set' + meshType[0] + measure[0] + '()')
                quality.Update()
                res += '\n%s\n%s' % (
                    measure[1],
                    DumpQualityStats(quality, 'Mesh ' + meshType[1] + ' Quality')
                )
            res += '\n'

    info = """\n\nDefinition of the different quality measures is given
    in the verdict library manual :
    http://www.vtk.org/Wiki/images/6/6b/VerdictManual-revA.pdf\n"""

    res += info
    print(res)
    return
